import argparse

import numpy as np
import wandb
from adaboost import AdaBoostF
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score
from sklearn.tree import DecisionTreeClassifier

from openfl.interface.interactive_api.experiment import FLExperiment, TaskInterface, ModelInterface
from openfl.interface.interactive_api.federation import Federation
from vowelDataset import vowelDataset

LOG_WANDB = False

parser = argparse.ArgumentParser(description="Script")
parser.add_argument("--rounds", default=3, type=int)
parser.add_argument("--seed", default=42, type=int)
parser.add_argument("--server", default='localhost', type=str, help="server address")
args = parser.parse_args()

random_state = np.random.RandomState(args.seed)

client_id = 'api'
cert_dir = '../cert'

task_interface = TaskInterface()


@task_interface.register_fl_task(model='model', data_loader='train_loader', device='device', optimizer='optimizer',
                                 adaboost_coeff='adaboost_coeff', name='name', nn=False)
def train_adaboost(model, train_loader, device, optimizer, adaboost_coeff, name):
    X, y = train_loader
    X, y = np.array(X), np.array(y)
    adaboost_coeff = np.array(adaboost_coeff)

    weak_learner = model.get(0)
    ids = np.random.choice(X.shape[0], size=X.shape[0], replace=True, p=adaboost_coeff / adaboost_coeff.sum())
    weak_learner.fit(X[ids], y[ids])

    pred = weak_learner.predict(X)
    metric = accuracy_score(y, pred)
    if LOG_WANDB:
        wandb.log({"weak_train_accuracy": accuracy_score(y, pred),
                   "weak_train_precision": precision_score(y, pred, average="micro"),
                   "weak_train_recall": recall_score(y, pred, average="micro"),
                   "weak_train_f1": f1_score(y, pred, average="micro")},
                  commit=False)
    return {'accuracy': metric}


@task_interface.register_fl_task(model='model', data_loader='val_loader', device='device',
                                 adaboost_coeff='adaboost_coeff', name='name', nn=False)
def validate_weak_learners(model, val_loader, device, adaboost_coeff, name):
    X, y = val_loader
    X, y = np.array(X), np.array(y)
    adaboost_coeff = np.array(adaboost_coeff)

    rank = int(str(name).split('_')[1]) - 1

    error = []
    miss = []
    for idx, weak_learner in enumerate(model.get_estimators()):
        pred = weak_learner.predict(X)
        mispredictions = y != pred
        error.append(sum(adaboost_coeff[mispredictions]))
        miss.append(mispredictions)
        if idx == rank:
            if LOG_WANDB:
                wandb.log({"weak_test_accuracy": accuracy_score(y, pred),
                           "weak_test_precision": precision_score(y, pred, average="micro"),
                           "weak_test_recall": recall_score(y, pred, average="micro"),
                           "weak_test_f1": f1_score(y, pred, average="micro")},
                          commit=False)
    # TODO: piccolo trick, alla fine di ogni vettore errori viene mandata la norma dei pesi locali
    error.append(sum(adaboost_coeff))

    return {'errors': error}, {'misprediction': miss}


@task_interface.register_fl_task(model='model', data_loader='val_loader', device='device', nn=False)
def adaboost_update(model, val_loader, device):
    return {'adaboost_update': 0}


@task_interface.register_fl_task(model='model', data_loader='val_loader', device='device', name='name', nn=False)
def validate_adaboost(model, val_loader, device, name):
    X, y = val_loader
    pred = model.predict(np.array(X))
    f1 = f1_score(y, pred, average="micro")

    if LOG_WANDB:
        wandb.log({"test_accuracy": accuracy_score(y, pred),
                   "test_precision": precision_score(y, pred, average="micro"),
                   "test_recall": recall_score(y, pred, average="micro"),
                   "test_f1": f1_score(y, pred, average="micro")})

    return {'F1 score': f1}


federation = Federation(client_id=client_id, director_node_fqdn=args.server, director_port='50052', tls=False)
fl_experiment = FLExperiment(federation=federation, experiment_name="AdaboostF_vowel",
                             serializer_plugin='openfl.plugins.interface_serializer.dill_serializer.DillSerializer',
                             load_default_plan=False, nn=False)
model_interface = ModelInterface(
    model=AdaBoostF(base_estimator=DecisionTreeClassifier(max_leaf_nodes=10), n_classes=11),
    optimizer=None,
    framework_plugin='openfl.plugins.frameworks_adapters.generic_adapter.GenericAdapter')
federated_dataset = vowelDataset()

fl_experiment.start(
    model_provider=model_interface,
    task_keeper=task_interface,
    data_loader=federated_dataset,
    rounds_to_train=args.rounds,
    opt_treatment='CONTINUE_GLOBAL'
)

fl_experiment.stream_metrics(tensorboard_logs=False)
fl_experiment.remove_experiment_data()



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Adam optimizer module."""

from typing import Dict
from typing import Optional
from typing import Tuple

import numpy as np

from .base_optimizer import Optimizer


class NumPyAdam(Optimizer):
    """Adam optimizer implementation.

    Original paper: https://openreview.net/forum?id=ryQu7f-RZ
    """

    def __init__(
        self,
        *,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        betas: Tuple[float, float] = (0.9, 0.999),
        initial_accumulator_value: float = 0.0,
        epsilon: float = 1e-8,
    ) -> None:
        """Initialize.

        Args:
            params: Parameters to be stored for optimization.
            model_interface: Model interface instance to provide parameters.
            learning_rate: Tuning parameter that determines
                the step size at each iteration.
            betas: Coefficients used for computing running
                averages of gradient and its square.
            initial_accumulator_value: Initial value for gradients
                and squared gradients.
            epsilon: Value for computational stability.
        """
        super().__init__()

        if model_interface is None and params is None:
            raise ValueError('Should provide one of the params or model_interface')

        if learning_rate < 0:
            raise ValueError(
                f'Invalid learning rate: {learning_rate}. Learning rate must be >= 0.')
        if not 0.0 <= betas[0] < 1:
            raise ValueError(
                f'Invalid betas[0] value: {betas[0]}. betas[0] must be in [0, 1).')
        if not 0.0 <= betas[1] < 1:
            raise ValueError(
                f'Invalid betas[1] value: {betas[1]}. betas[1] must be in [0, 1).')
        if initial_accumulator_value < 0:
            raise ValueError(
                f'Invalid initial_accumulator_value value: {initial_accumulator_value}. \
                Initial accumulator value must be >= 0.')
        if epsilon <= 0:
            raise ValueError(
                f'Invalid epsilon value: {epsilon}. Epsilon avalue must be > 0.')

        self.params = params

        if params is None and model_interface is not None:
            self._set_params_from_model(model_interface)

        self.learning_rate = learning_rate
        self.beta_1, self.beta_2 = betas
        self.initial_accumulator_value = initial_accumulator_value
        self.epsilon = epsilon
        self.current_step: Dict[str, int] = {param_name: 0 for param_name in self.params}

        self.grads_first_moment, self.grads_second_moment = {}, {}

        for param_name in self.params:
            self.grads_first_moment[param_name] = np.full_like(self.params[param_name],
                                                               self.initial_accumulator_value)
            self.grads_second_moment[param_name] = np.full_like(self.params[param_name],
                                                                self.initial_accumulator_value)

    def _update_first_moment(self, grad_name: str, grad: np.ndarray) -> None:
        """Update gradients first moment."""
        self.grads_first_moment[grad_name] = (self.beta_1
                                              * self.grads_first_moment[grad_name]
                                              + ((1.0 - self.beta_1) * grad))

    def _update_second_moment(self, grad_name: str, grad: np.ndarray) -> None:
        """Update gradients second moment."""
        self.grads_second_moment[grad_name] = (self.beta_2
                                               * self.grads_second_moment[grad_name]
                                               + ((1.0 - self.beta_2) * grad**2))

    def step(self, gradients: Dict[str, np.ndarray]) -> None:
        """
        Perform a single step for parameter update.

        Implement Adam optimizer weights update rule.

        Args:
            gradients: Partial derivatives with respect to optimized parameters.
        """
        for grad_name in gradients:
            if grad_name not in self.grads_first_moment:
                raise KeyError(f"Key {grad_name} doesn't exist in optimized parameters")

            grad = gradients[grad_name]

            self._update_first_moment(grad_name, grad)
            self._update_second_moment(grad_name, grad)

            t = self.current_step[grad_name] + 1
            mean = self.grads_first_moment[grad_name]
            var = self.grads_second_moment[grad_name]

            grads_first_moment_normalized = mean / (1. - self.beta_1 ** t)
            grads_second_moment_normalized = var / (1. - self.beta_2 ** t)

            # Make an update for a group of parameters
            self.params[grad_name] -= (self.learning_rate * grads_first_moment_normalized
                                       / (np.sqrt(grads_second_moment_normalized) + self.epsilon))

            self.current_step[grad_name] += 1



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Compute ArcFace loss and Triplet loss."""

import math

import torch
import torch.nn.functional as F
from torch import nn


class ArcFaceLoss(nn.Module):
    """ArcFace loss."""

    def __init__(self, margin=0.1, scale=16, easy_margin=False):
        """Initialize ArcFace loss."""
        super(ArcFaceLoss, self).__init__()
        self.m = margin
        self.s = scale
        self.easy_margin = easy_margin

    def forward(self, pred, target):
        """Compute forward."""
        # make a one-hot index
        index = pred.data * 0.0  # size = (B, Classnum)
        index.scatter_(1, target.data.view(-1, 1), 1)
        index = index.bool()

        cos_m = math.cos(self.m)
        sin_m = math.sin(self.m)
        cos_t = pred[index]
        sin_t = torch.sqrt(1.0 - cos_t * cos_t)
        cos_t_add_m = cos_t * cos_m - sin_t * sin_m

        cond_v = cos_t - math.cos(math.pi - self.m)
        cond = F.relu(cond_v)
        keep = cos_t - math.sin(math.pi - self.m) * self.m

        cos_t_add_m = torch.where(cond.bool(), cos_t_add_m, keep)

        output = pred * 1.0  # size = (B, Classnum)
        output[index] = cos_t_add_m
        output = self.s * output

        return F.cross_entropy(output, target)


class TripletLoss(nn.Module):
    """
    Triplet loss with hard positive/negative mining.

    Reference:
    Hermans et al. In Defense of the Triplet Loss for Person Re-Identification. arXiv:1703.07737.

    Code imported from https://github.com/Cysu/open-reid/blob/master/reid/loss/triplet.py.

    Args:
        margin (float): margin for triplet.
        distance (str): distance for triplet.
    """

    def __init__(self, margin=0.3, distance='cosine'):
        """Initialize Triplet loss."""
        super(TripletLoss, self).__init__()

        self.distance = distance
        self.margin = margin
        self.ranking_loss = nn.MarginRankingLoss(margin=margin)

    def forward(self, inputs, targets):
        """
        Compute forward.

        Args:
            inputs: feature matrix with shape (batch_size, feat_dim)
            targets: ground truth labels with shape (num_classes)
        """
        n = inputs.size(0)

        # Compute pairwise distance, replace by the official when merged
        inputs = F.normalize(inputs, p=2, dim=1)
        dist = - torch.mm(inputs, inputs.t())

        # For each anchor, find the hardest positive and negative
        mask = targets.expand(n, n).eq(targets.expand(n, n).t())
        dist_ap, dist_an = [], []
        for i in range(n):
            dist_ap.append(dist[i][mask[i]].max().unsqueeze(0))
            dist_an.append(dist[i][mask[i] == 0].min().unsqueeze(0))
        dist_ap = torch.cat(dist_ap)
        dist_an = torch.cat(dist_an)

        # Compute ranking hinge loss
        y = torch.ones_like(dist_an)
        loss = self.ranking_loss(dist_an, dist_ap, y)

        return loss



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import tensorflow as tf


def dice_coef(target, prediction, axis=(1, 2, 3), smooth=0.0001):
    """
    Sorenson Dice.

    Returns
    -------
    dice coefficient (float)
    """
    prediction = tf.round(prediction)  # Round to 0 or 1

    intersection = tf.reduce_sum(target * prediction, axis=axis)
    union = tf.reduce_sum(target + prediction, axis=axis)
    numerator = tf.constant(2.) * intersection + smooth
    denominator = union + smooth
    coef = numerator / denominator

    return tf.reduce_mean(coef)


def soft_dice_coef(target, prediction, axis=(1, 2, 3), smooth=0.0001):
    """
    Soft Sorenson Dice.

    Does not round the predictions to either 0 or 1.

    Returns
    -------
    soft dice coefficient (float)
    """
    intersection = tf.reduce_sum(target * prediction, axis=axis)
    union = tf.reduce_sum(target + prediction, axis=axis)
    numerator = tf.constant(2.) * intersection + smooth
    denominator = union + smooth
    coef = numerator / denominator

    return tf.reduce_mean(coef)


def dice_loss(target, prediction, axis=(1, 2, 3), smooth=0.0001):
    """
    Sorenson (Soft) Dice loss.

    Using -log(Dice) as the loss since it is better behaved.
    Also, the log allows avoidance of the division which
    can help prevent underflow when the numbers are very small.

    Returns
    -------
    dice loss (float)
    """
    intersection = tf.reduce_sum(prediction * target, axis=axis)
    p = tf.reduce_sum(prediction, axis=axis)
    t = tf.reduce_sum(target, axis=axis)
    numerator = tf.reduce_mean(intersection + smooth)
    denominator = tf.reduce_mean(t + p + smooth)
    dice_loss = -tf.math.log(2. * numerator) + tf.math.log(denominator)

    return dice_loss


def build_model(input_shape,
                n_cl_out=1,
                use_upsampling=False,
                dropout=0.2,
                print_summary=True,
                seed=816,
                depth=5,
                dropout_at=(2, 3),
                initial_filters=16,
                batch_norm=True,
                **kwargs):
    """Build the TensorFlow model.

    Args:
        input_tensor: input shape ot the model
        use_upsampling (bool): True = use bilinear interpolation;
                            False = use transposed convolution (Default=False)
        n_cl_out (int): Number of channels in output layer (Default=1)
        dropout (float): Dropout percentage (Default=0.2)
        print_summary (bool): True = print the model summary (Default = True)
        seed: random seed (Default=816)
        depth (int): Number of max pooling layers in encoder (Default=5)
        dropout_at: Layers to perform dropout after (Default=[2,3])
        initial_filters (int): Number of filters in first convolutional
        layer (Default=16)
        batch_norm (bool): True = use batch normalization (Default=True)
        **kwargs: Additional parameters to pass to the function
    """
    if (input_shape[0] % (2**depth)) > 0:
        raise ValueError(f'Crop dimension must be a multiple of 2^(depth of U-Net) = {2**depth}')

    inputs = tf.keras.layers.Input(input_shape, name='brats_mr_image')

    activation = tf.keras.activations.relu

    params = {'kernel_size': (3, 3, 3), 'activation': activation,
              'padding': 'same',
              'kernel_initializer': tf.keras.initializers.he_uniform(seed=seed)}

    convb_layers = {}

    net = inputs
    filters = initial_filters
    for i in range(depth):
        name = f'conv{i + 1}a'
        net = tf.keras.layers.Conv3D(name=name, filters=filters, **params)(net)
        if i in dropout_at:
            net = tf.keras.layers.Dropout(dropout)(net)
        name = f'conv{i + 1}b'
        net = tf.keras.layers.Conv3D(name=name, filters=filters, **params)(net)
        if batch_norm:
            net = tf.keras.layers.BatchNormalization()(net)
        convb_layers[name] = net
        # only pool if not last level
        if i != depth - 1:
            name = f'pool{i + 1}'
            net = tf.keras.layers.MaxPooling3D(name=name, pool_size=(2, 2, 2))(net)
            filters *= 2

    # do the up levels
    filters //= 2
    for i in range(depth - 1):
        if use_upsampling:
            up = tf.keras.layers.UpSampling3D(
                name=f'up{depth + i + 1}', size=(2, 2, 2))(net)
        else:
            up = tf.keras.layers.Conv3DTranspose(name=f'transConv{depth + i + 1}',
                                                 filters=filters,
                                                 kernel_size=(2, 2, 2),
                                                 strides=(2, 2, 2),
                                                 padding='same')(net)
        net = tf.keras.layers.concatenate(
            [up, convb_layers[f'conv{depth - i - 1}b']],
            axis=-1
        )
        net = tf.keras.layers.Conv3D(
            name=f'conv{depth + i + 1}a',
            filters=filters, **params)(net)
        net = tf.keras.layers.Conv3D(
            name=f'conv{depth + i + 1}b',
            filters=filters, **params)(net)
        filters //= 2

    net = tf.keras.layers.Conv3D(name='prediction', filters=n_cl_out,
                                 kernel_size=(1, 1, 1),
                                 activation='sigmoid')(net)

    model = tf.keras.models.Model(inputs=[inputs], outputs=[net])

    return model



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Director module."""

import asyncio
import logging
import time
from collections import defaultdict
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Union

from .experiment import Experiment
from .experiment import ExperimentsRegistry
from .experiment import Status

logger = logging.getLogger(__name__)

ENVOY_HEALTH_CHECK_PERIOD = 60  # in seconds


class Director:
    """Director class."""

    def __init__(
            self, *,
            tls: bool = True,
            root_certificate: Union[Path, str] = None,
            private_key: Union[Path, str] = None,
            certificate: Union[Path, str] = None,
            sample_shape: list = None,
            target_shape: list = None,
            settings: dict = None
    ) -> None:
        """Initialize a director object."""
        self.sample_shape, self.target_shape = sample_shape, target_shape
        self._shard_registry = {}
        self.tls = tls
        self.root_certificate = root_certificate
        self.private_key = private_key
        self.certificate = certificate
        self.experiments_registry = ExperimentsRegistry()
        self.settings = settings or {}
        self.col_exp_queues = defaultdict(asyncio.Queue)
        self.col_exp = {}

    def acknowledge_shard(self, shard_info: dict) -> bool:
        """Save shard info to shard registry if it's acceptable."""
        is_accepted = False
        if (self.sample_shape != shard_info['sample_shape']
                or self.target_shape != shard_info['target_shape']):
            logger.info('Request was not accepted')
            return is_accepted
        logger.info('Request was accepted')
        hc_period = self.settings.get('envoy_health_check_period', ENVOY_HEALTH_CHECK_PERIOD)
        self._shard_registry[shard_info['node_info']['name']] = {
            'shard_info': shard_info,
            'is_online': True,
            'is_experiment_running': False,
            'valid_duration': 2 * hc_period,
            'last_updated': time.time(),
        }
        is_accepted = True
        return is_accepted

    async def set_new_experiment(
            self, *,
            experiment_name: str,
            sender_name: str,
            tensor_dict: dict,
            collaborator_names: Iterable[str],
            experiment_archive_path: Path,
    ) -> bool:
        """Set new experiment."""
        experiment = Experiment(
            name=experiment_name,
            archive_path=experiment_archive_path,
            collaborators=list(collaborator_names),
            users=[sender_name],
            sender=sender_name,
            init_tensor_dict=tensor_dict,
        )
        self.experiments_registry.add(experiment)
        return True

    def get_trained_model(self, experiment_name: str, caller: str, model_type: str):
        """Get trained model."""
        if (experiment_name not in self.experiments_registry
                or caller not in self.experiments_registry[experiment_name].users):
            logger.error('No experiment data in the stash')
            return None

        aggregator = self.experiments_registry[experiment_name].aggregator

        if aggregator.last_tensor_dict is None:
            logger.error('Aggregator have no aggregated model to return')
            return None

        if model_type == 'best':
            return aggregator.best_tensor_dict
        elif model_type == 'last':
            return aggregator.last_tensor_dict
        else:
            logger.error('Unknown model type required.')
            return None

    def get_experiment_data(self, experiment_name: str) -> Path:
        """Get experiment data."""
        return self.experiments_registry[experiment_name].archive_path

    async def wait_experiment(self, envoy_name: str) -> str:
        """Wait an experiment."""
        self.col_exp[envoy_name] = None
        queue = self.col_exp_queues[envoy_name]
        experiment_name = await queue.get()
        self.col_exp[envoy_name] = experiment_name

        return experiment_name

    def get_dataset_info(self):
        """Get dataset info."""
        return self.sample_shape, self.target_shape

    def get_registered_shards(self) -> list:  # Why is it here?
        """Get registered shard infos."""
        return [shard_status['shard_info'] for shard_status in self._shard_registry.values()]

    async def stream_metrics(self, experiment_name: str, caller: str):
        """
        Stream metrics from the aggregator.

        This method takes next metric dictionary from the aggregator's queue
        and returns it to the caller.

        Inputs:
            experiment_name - string id for experiment
            caller - string id for experiment owner

        Returns:
            metric_dict - {'metric_origin','task_name','metric_name','metric_value','round'}
                if the queue is not empty
            None - f queue is empty but the experiment is still running

        Raises:
            StopIteration - if the experiment is finished and there is no more metrics to report
        """
        if (experiment_name not in self.experiments_registry
                or caller not in self.experiments_registry[experiment_name].users):
            raise Exception(
                f'No experiment name "{experiment_name}" in experiments list, or caller "{caller}"'
                f' does not have access to this experiment'
            )

        while not self.experiments_registry[experiment_name].aggregator:
            await asyncio.sleep(1)
        aggregator = self.experiments_registry[experiment_name].aggregator

        while True:
            if not aggregator.metric_queue.empty():
                yield aggregator.metric_queue.get()
                continue

            if aggregator.all_quit_jobs_sent() and aggregator.metric_queue.empty():
                return

            yield None

    def remove_experiment_data(self, experiment_name: str, caller: str):
        """Remove experiment data from stash."""
        if (experiment_name in self.experiments_registry
                and caller in self.experiments_registry[experiment_name].users):
            self.experiments_registry.remove(experiment_name)

    def set_experiment_failed(self, *, experiment_name: str, collaborator_name: str):
        """Set experiment failed."""
        if experiment_name in self.experiments_registry:
            aggregator = self.experiments_registry[experiment_name].aggregator
            aggregator.stop(failed_collaborator=collaborator_name)

    def update_envoy_status(
            self, *,
            envoy_name: str,
            is_experiment_running: bool,
            cuda_devices_status: list = None,
    ) -> int:
        """Accept health check from envoy."""
        shard_info = self._shard_registry.get(envoy_name)
        if not shard_info:
            raise Exception(f'Unknown shard {envoy_name}')

        hc_period = self.settings.get('envoy_health_check_period', ENVOY_HEALTH_CHECK_PERIOD)
        shard_info['is_online']: True
        shard_info['is_experiment_running'] = is_experiment_running
        shard_info['valid_duration'] = 2 * hc_period
        shard_info['last_updated'] = time.time()

        if cuda_devices_status is not None:
            for i in range(len(cuda_devices_status)):
                shard_info['shard_info']['node_info']['cuda_devices'][i] = cuda_devices_status[i]

        return hc_period

    def get_envoys(self) -> list:
        """Get a status information about envoys."""
        logger.info(f'Shard registry: {self._shard_registry}')
        for envoy_info in self._shard_registry.values():
            envoy_info['is_online'] = (
                time.time() < envoy_info.get('last_updated', 0)
                + envoy_info.get('valid_duration', 0)
            )
            envoy_name = envoy_info['shard_info']['node_info']['name']
            envoy_info['experiment_name'] = self.col_exp[envoy_name]

        return self._shard_registry.values()

    def get_experiments_list(self, caller: str) -> list:
        """Get experiments list for specific user."""
        experiments = self.experiments_registry.get_user_experiments(caller)
        result = []
        for exp in experiments:
            exp_data = {
                'name': exp.name,
                'status': exp.status,
                'collaborators_amount': len(exp.collaborators),
            }
            progress = _get_experiment_progress(exp)
            if progress is not None:
                exp_data['progress'] = progress
            if exp.aggregator:
                tasks_amount = len({
                    task['function']
                    for task in exp.aggregator.assigner.tasks.values()
                })
                exp_data['tasks_amount'] = tasks_amount
            result.append(exp_data)

        return result

    def get_experiment_description(self, caller: str, name: str) -> dict:
        """Get a experiment information by name for specific user."""
        exp = self.experiments_registry.get(name)
        if not exp or caller not in exp.users:
            return {}
        progress = _get_experiment_progress(exp)
        model_statuses = _get_model_download_statuses(exp)
        tasks = _get_experiment_tasks(exp)
        collaborators = _get_experiment_collaborators(exp)
        result = {
            'name': name,
            'status': exp.status,
            'current_round': exp.aggregator.round_number,
            'total_rounds': exp.aggregator.rounds_to_train,
            'download_statuses': {
                'models': model_statuses,
                'logs': [{
                    'name': 'aggregator',
                    'status': 'ready'
                }],
            },
            'collaborators': collaborators,
            'tasks': tasks,
            'progress': progress
        }
        return result

    async def start_experiment_execution_loop(self):
        """Run task to monitor and run experiments."""
        while True:
            async with self.experiments_registry.get_next_experiment() as experiment:
                loop = asyncio.get_event_loop()
                run_aggregator_future = loop.create_task(experiment.start(
                    root_certificate=self.root_certificate,
                    certificate=self.certificate,
                    private_key=self.private_key,
                    tls=self.tls,
                ))
                for col_name in experiment.collaborators:
                    queue = self.col_exp_queues[col_name]
                    await queue.put(experiment.name)
                await run_aggregator_future


def _get_model_download_statuses(experiment) -> List[dict]:
    best_model_status = 'ready' if experiment.aggregator.best_tensor_dict else 'pending'
    last_model_status = 'ready' if experiment.aggregator.last_tensor_dict else 'pending'
    model_statuses = [{
        'name': 'best',
        'status': best_model_status,
    }, {
        'name': 'last',
        'status': last_model_status,
    }, {
        'name': 'init',
        'status': 'ready'
    }]
    return model_statuses


def _get_experiment_progress(experiment) -> Union[float, None]:
    if experiment.status == Status.IN_PROGRESS:
        return experiment.aggregator.round_number / experiment.aggregator.rounds_to_train


def _get_experiment_tasks(experiment) -> List[dict]:
    return [{
        'name': task['function'],
        'description': 'Task description Mock',
    } for task in experiment.aggregator.assigner.tasks.values()]


def _get_experiment_collaborators(experiment) -> List[dict]:
    return [{
        'name': name,
        'status': 'pending_mock',
        'progress': 0.0,
        'round': 0,
        'current_task': 'Current Task Mock',
        'next_task': 'Next Task Mock'
    } for name in experiment.aggregator.authorized_cols]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openfl/protocols/aggregator.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from openfl.protocols import base_pb2 as openfl_dot_protocols_dot_base__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!openfl/protocols/aggregator.proto\x12\x11openfl.aggregator\x1a\x1bopenfl/protocols/base.proto\"o\n\rMessageHeader\x12\x0e\n\x06sender\x18\x01 \x01(\t\x12\x10\n\x08receiver\x18\x02 \x01(\t\x12\x17\n\x0f\x66\x65\x64\x65ration_uuid\x18\x03 \x01(\t\x12#\n\x1bsingle_col_cert_common_name\x18\x04 \x01(\t\"C\n\x0fGetTasksRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\"S\n\x04Task\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rfunction_name\x18\x02 \x01(\t\x12\x11\n\ttask_type\x18\x03 \x01(\t\x12\x13\n\x0b\x61pply_local\x18\x04 \x01(\x08\"\xa4\x01\n\x10GetTasksResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12&\n\x05tasks\x18\x03 \x03(\x0b\x32\x17.openfl.aggregator.Task\x12\x12\n\nsleep_time\x18\x04 \x01(\x05\x12\x0c\n\x04quit\x18\x05 \x01(\x08\"\xbb\x01\n\x10GetTensorRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x13\n\x0btensor_name\x18\x02 \x01(\t\x12\x14\n\x0cround_number\x18\x03 \x01(\x05\x12\x0e\n\x06report\x18\x04 \x01(\x08\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12\x18\n\x10require_lossless\x18\x06 \x01(\x08\x12\x12\n\naggregated\x18\x07 \x01(\x08\"y\n\x11GetTensorResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x1c\n\x06tensor\x18\x03 \x01(\x0b\x32\x0c.NamedTensor\"\x9a\x01\n\x0bTaskResults\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cround_number\x18\x02 \x01(\x05\x12\x11\n\ttask_name\x18\x03 \x01(\t\x12\x11\n\tdata_size\x18\x04 \x01(\x05\x12\x1d\n\x07tensors\x18\x05 \x03(\x0b\x32\x0c.NamedTensor\"P\n\x1cSendLocalTaskResultsResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\"1\n\x16GetMetricStreamRequest\x12\x17\n\x0f\x65xperiment_name\x18\x01 \x01(\t\"}\n\x17GetMetricStreamResponse\x12\x15\n\rmetric_origin\x18\x01 \x01(\t\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x13\n\x0bmetric_name\x18\x03 \x01(\t\x12\x14\n\x0cmetric_value\x18\x04 \x01(\x02\x12\r\n\x05round\x18\x05 \x01(\r\"\xa7\x01\n\x16GetTrainedModelRequest\x12\x17\n\x0f\x65xperiment_name\x18\x02 \x01(\t\x12G\n\nmodel_type\x18\x03 \x01(\x0e\x32\x33.openfl.aggregator.GetTrainedModelRequest.ModelType\"+\n\tModelType\x12\x0e\n\nBEST_MODEL\x10\x00\x12\x0e\n\nLAST_MODEL\x10\x01\"8\n\x14TrainedModelResponse\x12 \n\x0bmodel_proto\x18\x01 \x01(\x0b\x32\x0b.ModelProto\"/\n\x1fGetExperimentDescriptionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"N\n GetExperimentDescriptionResponse\x12*\n\nexperiment\x18\x01 \x01(\x0b\x32\x16.ExperimentDescription\"i\n\x0cSynchRequest\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x11\n\ttask_name\x18\x02 \x01(\t\x12\x14\n\x0cround_number\x18\x03 \x01(\x05\"W\n\rSynchResponse\x12\x30\n\x06header\x18\x01 \x01(\x0b\x32 .openfl.aggregator.MessageHeader\x12\x14\n\x0cis_completed\x18\x02 \x01(\x08\x32\xc7\x05\n\nAggregator\x12U\n\x08GetTasks\x12\".openfl.aggregator.GetTasksRequest\x1a#.openfl.aggregator.GetTasksResponse\"\x00\x12O\n\x08GetSynch\x12\x1f.openfl.aggregator.SynchRequest\x1a .openfl.aggregator.SynchResponse\"\x00\x12X\n\tGetTensor\x12#.openfl.aggregator.GetTensorRequest\x1a$.openfl.aggregator.GetTensorResponse\"\x00\x12X\n\x14SendLocalTaskResults\x12\x0b.DataStream\x1a/.openfl.aggregator.SendLocalTaskResultsResponse\"\x00(\x01\x12l\n\x0fGetMetricStream\x12).openfl.aggregator.GetMetricStreamRequest\x1a*.openfl.aggregator.GetMetricStreamResponse\"\x00\x30\x01\x12g\n\x0fGetTrainedModel\x12).openfl.aggregator.GetTrainedModelRequest\x1a\'.openfl.aggregator.TrainedModelResponse\"\x00\x12\x85\x01\n\x18GetExperimentDescription\x12\x32.openfl.aggregator.GetExperimentDescriptionRequest\x1a\x33.openfl.aggregator.GetExperimentDescriptionResponse\"\x00\x62\x06proto3')



_MESSAGEHEADER = DESCRIPTOR.message_types_by_name['MessageHeader']
_GETTASKSREQUEST = DESCRIPTOR.message_types_by_name['GetTasksRequest']
_TASK = DESCRIPTOR.message_types_by_name['Task']
_GETTASKSRESPONSE = DESCRIPTOR.message_types_by_name['GetTasksResponse']
_GETTENSORREQUEST = DESCRIPTOR.message_types_by_name['GetTensorRequest']
_GETTENSORRESPONSE = DESCRIPTOR.message_types_by_name['GetTensorResponse']
_TASKRESULTS = DESCRIPTOR.message_types_by_name['TaskResults']
_SENDLOCALTASKRESULTSRESPONSE = DESCRIPTOR.message_types_by_name['SendLocalTaskResultsResponse']
_GETMETRICSTREAMREQUEST = DESCRIPTOR.message_types_by_name['GetMetricStreamRequest']
_GETMETRICSTREAMRESPONSE = DESCRIPTOR.message_types_by_name['GetMetricStreamResponse']
_GETTRAINEDMODELREQUEST = DESCRIPTOR.message_types_by_name['GetTrainedModelRequest']
_TRAINEDMODELRESPONSE = DESCRIPTOR.message_types_by_name['TrainedModelResponse']
_GETEXPERIMENTDESCRIPTIONREQUEST = DESCRIPTOR.message_types_by_name['GetExperimentDescriptionRequest']
_GETEXPERIMENTDESCRIPTIONRESPONSE = DESCRIPTOR.message_types_by_name['GetExperimentDescriptionResponse']
_SYNCHREQUEST = DESCRIPTOR.message_types_by_name['SynchRequest']
_SYNCHRESPONSE = DESCRIPTOR.message_types_by_name['SynchResponse']
_GETTRAINEDMODELREQUEST_MODELTYPE = _GETTRAINEDMODELREQUEST.enum_types_by_name['ModelType']
MessageHeader = _reflection.GeneratedProtocolMessageType('MessageHeader', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGEHEADER,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.MessageHeader)
  })
_sym_db.RegisterMessage(MessageHeader)

GetTasksRequest = _reflection.GeneratedProtocolMessageType('GetTasksRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTASKSREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTasksRequest)
  })
_sym_db.RegisterMessage(GetTasksRequest)

Task = _reflection.GeneratedProtocolMessageType('Task', (_message.Message,), {
  'DESCRIPTOR' : _TASK,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.Task)
  })
_sym_db.RegisterMessage(Task)

GetTasksResponse = _reflection.GeneratedProtocolMessageType('GetTasksResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTASKSRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTasksResponse)
  })
_sym_db.RegisterMessage(GetTasksResponse)

GetTensorRequest = _reflection.GeneratedProtocolMessageType('GetTensorRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTENSORREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTensorRequest)
  })
_sym_db.RegisterMessage(GetTensorRequest)

GetTensorResponse = _reflection.GeneratedProtocolMessageType('GetTensorResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETTENSORRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTensorResponse)
  })
_sym_db.RegisterMessage(GetTensorResponse)

TaskResults = _reflection.GeneratedProtocolMessageType('TaskResults', (_message.Message,), {
  'DESCRIPTOR' : _TASKRESULTS,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.TaskResults)
  })
_sym_db.RegisterMessage(TaskResults)

SendLocalTaskResultsResponse = _reflection.GeneratedProtocolMessageType('SendLocalTaskResultsResponse', (_message.Message,), {
  'DESCRIPTOR' : _SENDLOCALTASKRESULTSRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.SendLocalTaskResultsResponse)
  })
_sym_db.RegisterMessage(SendLocalTaskResultsResponse)

GetMetricStreamRequest = _reflection.GeneratedProtocolMessageType('GetMetricStreamRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICSTREAMREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetMetricStreamRequest)
  })
_sym_db.RegisterMessage(GetMetricStreamRequest)

GetMetricStreamResponse = _reflection.GeneratedProtocolMessageType('GetMetricStreamResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICSTREAMRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetMetricStreamResponse)
  })
_sym_db.RegisterMessage(GetMetricStreamResponse)

GetTrainedModelRequest = _reflection.GeneratedProtocolMessageType('GetTrainedModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTRAINEDMODELREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetTrainedModelRequest)
  })
_sym_db.RegisterMessage(GetTrainedModelRequest)

TrainedModelResponse = _reflection.GeneratedProtocolMessageType('TrainedModelResponse', (_message.Message,), {
  'DESCRIPTOR' : _TRAINEDMODELRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.TrainedModelResponse)
  })
_sym_db.RegisterMessage(TrainedModelResponse)

GetExperimentDescriptionRequest = _reflection.GeneratedProtocolMessageType('GetExperimentDescriptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDESCRIPTIONREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetExperimentDescriptionRequest)
  })
_sym_db.RegisterMessage(GetExperimentDescriptionRequest)

GetExperimentDescriptionResponse = _reflection.GeneratedProtocolMessageType('GetExperimentDescriptionResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTDESCRIPTIONRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.GetExperimentDescriptionResponse)
  })
_sym_db.RegisterMessage(GetExperimentDescriptionResponse)

SynchRequest = _reflection.GeneratedProtocolMessageType('SynchRequest', (_message.Message,), {
  'DESCRIPTOR' : _SYNCHREQUEST,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.SynchRequest)
  })
_sym_db.RegisterMessage(SynchRequest)

SynchResponse = _reflection.GeneratedProtocolMessageType('SynchResponse', (_message.Message,), {
  'DESCRIPTOR' : _SYNCHRESPONSE,
  '__module__' : 'openfl.protocols.aggregator_pb2'
  # @@protoc_insertion_point(class_scope:openfl.aggregator.SynchResponse)
  })
_sym_db.RegisterMessage(SynchResponse)

_AGGREGATOR = DESCRIPTOR.services_by_name['Aggregator']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGEHEADER._serialized_start=85
  _MESSAGEHEADER._serialized_end=196
  _GETTASKSREQUEST._serialized_start=198
  _GETTASKSREQUEST._serialized_end=265
  _TASK._serialized_start=267
  _TASK._serialized_end=350
  _GETTASKSRESPONSE._serialized_start=353
  _GETTASKSRESPONSE._serialized_end=517
  _GETTENSORREQUEST._serialized_start=520
  _GETTENSORREQUEST._serialized_end=707
  _GETTENSORRESPONSE._serialized_start=709
  _GETTENSORRESPONSE._serialized_end=830
  _TASKRESULTS._serialized_start=833
  _TASKRESULTS._serialized_end=987
  _SENDLOCALTASKRESULTSRESPONSE._serialized_start=989
  _SENDLOCALTASKRESULTSRESPONSE._serialized_end=1069
  _GETMETRICSTREAMREQUEST._serialized_start=1071
  _GETMETRICSTREAMREQUEST._serialized_end=1120
  _GETMETRICSTREAMRESPONSE._serialized_start=1122
  _GETMETRICSTREAMRESPONSE._serialized_end=1247
  _GETTRAINEDMODELREQUEST._serialized_start=1250
  _GETTRAINEDMODELREQUEST._serialized_end=1417
  _GETTRAINEDMODELREQUEST_MODELTYPE._serialized_start=1374
  _GETTRAINEDMODELREQUEST_MODELTYPE._serialized_end=1417
  _TRAINEDMODELRESPONSE._serialized_start=1419
  _TRAINEDMODELRESPONSE._serialized_end=1475
  _GETEXPERIMENTDESCRIPTIONREQUEST._serialized_start=1477
  _GETEXPERIMENTDESCRIPTIONREQUEST._serialized_end=1524
  _GETEXPERIMENTDESCRIPTIONRESPONSE._serialized_start=1526
  _GETEXPERIMENTDESCRIPTIONRESPONSE._serialized_end=1604
  _SYNCHREQUEST._serialized_start=1606
  _SYNCHREQUEST._serialized_end=1711
  _SYNCHRESPONSE._serialized_start=1713
  _SYNCHRESPONSE._serialized_end=1800
  _AGGREGATOR._serialized_start=1803
  _AGGREGATOR._serialized_end=2514
# @@protoc_insertion_point(module_scope)



"""Copyright (C) 2020-2021 Intel Corporation
   SPDX-License-Identifier: Apache-2.0

Licensed subject to the terms of the separately executed evaluation
license agreement between Intel Corporation and you.
"""
from tensorflow import keras

from openfl.federated import KerasTaskRunner


def build_model(latent_dim, num_encoder_tokens, num_decoder_tokens, **kwargs):
    """
    Define the model architecture.

    Args:
        input_shape (numpy.ndarray): The shape of the data
        num_classes (int): The number of classes of the dataset
    Returns:
        tensorflow.python.keras.engine.sequential.Sequential: The model defined in Keras
    """
    encoder_inputs = keras.Input(shape=(None, num_encoder_tokens))
    encoder = keras.layers.LSTM(latent_dim, return_state=True)
    encoder_outputs, state_h, state_c = encoder(encoder_inputs)

    # We discard `encoder_outputs` and only keep the states.
    encoder_states = [state_h, state_c]

    # Set up the decoder, using `encoder_states` as initial state.
    decoder_inputs = keras.Input(shape=(None, num_decoder_tokens))

    # We set up our decoder to return full output sequences,
    # and to return internal states as well. We don't use the
    # return states in the training model, but we will use them in inference.
    decoder_lstm = keras.layers.LSTM(latent_dim, return_sequences=True, return_state=True)
    decoder_outputs, _, _ = decoder_lstm(decoder_inputs, initial_state=encoder_states)
    decoder_dense = keras.layers.Dense(num_decoder_tokens, activation='softmax')
    decoder_outputs = decoder_dense(decoder_outputs)

    # Define the model that will turn
    # `encoder_input_data` & `decoder_input_data` into `decoder_target_data`
    model = keras.Model([encoder_inputs, decoder_inputs], decoder_outputs)

    model.compile(
        optimizer='rmsprop', loss='categorical_crossentropy', metrics=['accuracy']
    )

    return model


class KerasNLP(KerasTaskRunner):
    """A basic convolutional neural network model."""

    def __init__(self, latent_dim, **kwargs):
        """
        Init taskrunner.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        super().__init__(**kwargs)

        self.model = build_model(latent_dim,
                                 self.data_loader.num_encoder_tokens,
                                 self.data_loader.num_decoder_tokens,
                                 **kwargs)

        self.initialize_tensorkeys_for_functions()

        self.model.summary(print_fn=self.logger.info)

        if self.data_loader is not None:
            self.logger.info(f'Train Set Size : {self.get_train_data_size()}')



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""PyTorch optimizers package."""
import pkgutil

if pkgutil.find_loader('torch'):
    from .fedprox import FedProxOptimizer # NOQA
    from .fedprox import FedProxAdam # NOQA



# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Custom model numpy adapter."""

from openfl.plugins.frameworks_adapters.framework_adapter_interface import (
    FrameworkAdapterPluginInterface,
)


class CustomFrameworkAdapter(FrameworkAdapterPluginInterface):
    """Framework adapter plugin class."""

    @staticmethod
    def get_tensor_dict(model, optimizer=None):
        """Extract tensors from a model."""
        return {'w': model.weights}

    @staticmethod
    def set_tensor_dict(model, tensor_dict, optimizer=None, device='cpu'):
        """Load tensors to a model."""
        model.weights = tensor_dict['w']



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import tqdm

from openfl.federated import PyTorchTaskRunner
from openfl.utilities import TensorKey


def cross_entropy(output, target):
    """Calculate Cross-entropy loss."""
    return F.cross_entropy(input=output, target=target)


class PyTorchCNN(PyTorchTaskRunner):
    """Simple CNN for classification."""

    def __init__(self, **kwargs):
        """Initialize.

        Args:
            **kwargs: Additional arguments to pass to the function
        """
        super().__init__(loss_fn=cross_entropy, **kwargs)

        torch.manual_seed(0)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

        self.num_classes = self.data_loader.num_classes
        self.init_network(device=self.device, **kwargs)
        self._init_optimizer(lr=kwargs.get('lr'))
        self.initialize_tensorkeys_for_functions()

    def _init_optimizer(self, lr):
        """Initialize the optimizer."""
        self.optimizer = optim.Adam(self.parameters(), lr=float(lr or 1e-3))

    def init_network(self,
                     device,
                     print_model=True,
                     **kwargs):
        """Create the network (model).

        Args:
            device: The hardware device to use for training
            print_model (bool): Print the model topology (Default=True)
            **kwargs: Additional arguments to pass to the function

        """
        channel = self.data_loader.get_feature_shape()[
            0]  # (channel, dim1, dim2)
        conv_kwargs = {'kernel_size': 3, 'stride': 1, 'padding': 1}
        self.conv1 = nn.Conv2d(channel, 16, **conv_kwargs)
        self.conv2 = nn.Conv2d(16, 32, **conv_kwargs)
        self.conv3 = nn.Conv2d(32, 64, **conv_kwargs)
        self.conv4 = nn.Conv2d(64, 128, **conv_kwargs)
        self.conv5 = nn.Conv2d(128 + 32, 256, **conv_kwargs)
        self.conv6 = nn.Conv2d(256, 512, **conv_kwargs)
        self.conv7 = nn.Conv2d(512 + 128 + 32, 256, **conv_kwargs)
        self.conv8 = nn.Conv2d(256, 512, **conv_kwargs)
        self.fc1 = nn.Linear(1184 * 9 * 9, 128)
        self.fc2 = nn.Linear(128, 8)
        if print_model:
            print(self)
        self.to(device)

    def forward(self, x):
        """Forward pass of the model.

        Args:
            x: Data input to the model for the forward pass
        """
        x = F.relu(self.conv1(x))
        x = F.relu(self.conv2(x))
        maxpool = F.max_pool2d(x, 2, 2)

        x = F.relu(self.conv3(maxpool))
        x = F.relu(self.conv4(x))
        concat = torch.cat([maxpool, x], dim=1)
        maxpool = F.max_pool2d(concat, 2, 2)

        x = F.relu(self.conv5(maxpool))
        x = F.relu(self.conv6(x))
        concat = torch.cat([maxpool, x], dim=1)
        maxpool = F.max_pool2d(concat, 2, 2)

        x = F.relu(self.conv7(maxpool))
        x = F.relu(self.conv8(x))
        concat = torch.cat([maxpool, x], dim=1)
        maxpool = F.max_pool2d(concat, 2, 2)

        x = maxpool.flatten(start_dim=1)
        x = F.dropout(self.fc1(x), p=0.5)
        x = self.fc2(x)
        return x

    def validate(self, col_name, round_num, input_tensor_dict,
                 use_tqdm=False, **kwargs):
        """Validate.

        Run validation of the model on the local data.

        Args:
            col_name:            Name of the collaborator
            round_num:           What round is it
            input_tensor_dict:   Required input tensors (for model)
            use_tqdm (bool):     Use tqdm to print a progress
                                 bar (Default=True)

        Returns:
            global_output_dict:  Tensors to send back to the aggregator
            local_output_dict:   Tensors to maintain in the local TensorDB

        """
        self.rebuild_model(round_num, input_tensor_dict, validation=True)
        self.eval()
        val_score = 0
        total_samples = 0

        loader = self.data_loader.get_valid_loader()
        if use_tqdm:
            loader = tqdm.tqdm(loader, desc='validate')

        with torch.no_grad():
            for data, target in loader:
                samples = target.shape[0]
                total_samples += samples
                data, target = (torch.tensor(data).to(self.device),
                                torch.tensor(target).to(self.device))
                output = self(data)
                # get the index of the max log-probability
                pred = output.argmax(dim=1)
                val_score += pred.eq(target).sum().cpu().numpy()

        origin = col_name
        suffix = 'validate'
        if kwargs['apply'] == 'local':
            suffix += '_local'
        else:
            suffix += '_agg'
        tags = ('metric', suffix)
        # TODO figure out a better way to pass in metric for
        #  this pytorch validate function
        output_tensor_dict = {
            TensorKey('acc', origin, round_num, True, tags):
                np.array(val_score / total_samples)
        }

        # empty list represents metrics that should only be stored locally
        return output_tensor_dict, {}

    def reset_opt_vars(self):
        """Reset optimizer variables.

        Resets the optimizer state variables.

        """
        self._init_optimizer(lr=self.optimizer.defaults.get('lr'))
